"""
Module `filters` defines placeholders and how names are filtered.
"""
from .defaults import default_filename_template
from .defaults import default_latest_filename_template
from .defaults import load_placeholder, load_alias
from semantic_version import Version
import requests
import json

import re
from string import Template

from typing import Mapping, Optional, Callable

VERSION_REGEX = re.compile(
    r'v?(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(-(?P<status>\w+))?')
SPECIAL_VERSION_NAMES = ["latest", "nightly", "stable"]


def is_version(version):
    if version in SPECIAL_VERSION_NAMES:
        return True
    else:
        return bool(VERSION_REGEX.match(version))


def canonicalize_arch(arch: Optional[str]):
    """
        Canonicalize arch names to arch names defined by Julia schema.

        For example, `--sys=win` is allowed in jill but this is
        not listed in Julia's `versions-schema.json`. This function
        maps these alias into the legal names defined by
        `versions-schema.json`.

        The canonicalize rule is defined in `jill/config/alias.json`, for
        names not in that list, it's an identity map.
    """
    if arch is not None:
        return load_alias()["Arch"].get(arch.lower(), arch.lower())


def canonicalize_sys(os: Optional[str]):
    """
        Canonicalize sys names to OS names defined by Julia schema.

        For example, `--sys=win` is allowed in jill but this is
        not listed in Julia's `versions-schema.json`. This function
        maps these alias into the legal names defined by
        `versions-schema.json`.

        The canonicalize rule is defined in `jill/config/alias.json`, for
        names not in that list, it's an identity map.
    """
    if os is not None:
        return load_alias()["OS"].get(os.lower(), os.lower())


def identity(*args):
    return args[0] if len(args) == 1 else args


def no_validate(*args, **kwargs):
    return True


def _Osarch(os, arch):
    if os in ["win", "mac"]:
        return f_osarch(os, arch).capitalize()
    os, arch = f_osarch(os, arch).split('-')
    return os.capitalize() + '-' + arch


def _OSarch(os, arch):
    if os in ["win", "mac"]:
        return f_osarch(os, arch).upper()
    os, arch = f_osarch(os, arch).split('-')
    return os.upper() + '-' + arch


class NameFilter:
    def __init__(self,
                 name: str,
                 f: Callable = identity,
                 rules: Optional[Mapping] = None,
                 validate: Callable = no_validate):
        self.f = f
        self.name = name
        if rules:
            self.rules = rules
        else:
            rules = load_placeholder().get(name, dict())
            self.rules = rules if rules else dict()
        self.validate = validate

    def __call__(self, *args, **kwargs):
        if not self.validate(*args, **kwargs):
            # TODO: add error handler
            msg = f"validation on {self.name} fails:\n"
            msg += f"  - args: {args}\n  - kwargs: {kwargs}\n"
            msg += f"Please check if you have passed the right parameters"
            raise ValueError(msg)
        # directly return rst if there're no special filter rules
        rst = self.f(*args, **kwargs)
        return self.rules.get(rst, rst)


f_major_version = NameFilter("version", lambda x: x.lstrip('v').split('.')[0],
                             validate=is_version)
f_minor_version = NameFilter("version", lambda x: '.'.join(x.lstrip('v').
                                                           split('.')[0:2]),
                             validate=is_version)
f_patch_version = NameFilter("version", lambda x: x.lstrip('v').split('-')[0],
                             validate=is_version)


def _vmajor_version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return 'v'+f_major_version(ver)


def _vminor_version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return 'v'+f_minor_version(ver)


def _vpatch_version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return 'v'+f_patch_version(ver)


def _version(ver):
    if ver in SPECIAL_VERSION_NAMES:
        return ver
    else:
        return ver.lstrip('v')


f_vmajor_version = NameFilter("version", _vmajor_version)
f_Vmajor_version = NameFilter(
    "version", lambda x: f_vmajor_version(x).capitalize())
f_vminor_version = NameFilter("version", _vminor_version)
f_Vminor_version = NameFilter(
    "version", lambda x: f_vminor_version(x).capitalize())
f_vpatch_version = NameFilter("version", _vpatch_version)
f_Vpatch_version = NameFilter(
    "version", lambda x: f_vpatch_version(x).capitalize())

f_version = NameFilter("version", _version, validate=is_version)


def _build(ver, cache={}):
    if ver in SPECIAL_VERSION_NAMES:
        return ''
    build = Version(ver).build
    if not build:
        return ''
    build = build[0]
    if len(build) >= 10:
        return build[0:10]
    # If its length is not 10, then we query the github API see what's the actual one
    # To reduce repeated queries, we store it in a dictionary
    # Example: https://s3.amazonaws.com/julialangnightlies/pretesting/mac/x64/1.8/julia-9cad1e0af8-mac64.tar.gz
    if build in cache:
        return cache[build]
    try:
        github_api = f"https://api.github.com/repos/julialang/julia/commits/{build}"
        data = json.loads(requests.get(github_api).content)
        cache[build] = data['sha'][0:10]
        return cache[build]
    except:
        return build


f_build = NameFilter("build", _build)

f_system = NameFilter("system")
f_System = NameFilter("system", f=lambda x: f_system(x).capitalize())
f_SYSTEM = NameFilter("system", f=lambda x: f_system(x).upper())

f_sys = NameFilter("sys")
f_Sys = NameFilter("sys", f=lambda x: f_sys(x).capitalize())
f_SYS = NameFilter("sys", f=lambda x: f_sys(x).upper())

f_os = NameFilter("os")
f_Os = NameFilter("os", f=lambda x: f_os(x).capitalize())
f_OS = NameFilter("os", f=lambda x: f_os(x).upper())

f_arch = NameFilter("arch")
f_Arch = NameFilter("arch", f=lambda x: f_arch(x).capitalize())
f_ARCH = NameFilter("arch", f=lambda x: f_arch(x).upper())

f_osarch = NameFilter("osarch", f=lambda os, arch: f"{os}-{arch}")
f_Osarch = NameFilter("osarch", f=_Osarch)
f_OSarch = NameFilter("osarch", f=_OSarch)

f_osbit = NameFilter("osbit", f=lambda os, arch: f"{os}{arch}")
f_bit = NameFilter("bit")
f_extension = NameFilter("extension")


def _meta_filename(t, *args, **kwargs):
    if not isinstance(t, Template):
        t = Template(t)
    return t.substitute(*args, **kwargs)


def _filename(*args, **kwargs):
    return _meta_filename(default_filename_template, *args, **kwargs)


def _latest_filename(**kwargs):
    return _meta_filename(default_latest_filename_template, **kwargs)


f_filename = NameFilter("filename", _filename)
f_latest_filename = NameFilter("filename", _latest_filename)


def generate_info(plain_version: str,
                  system: str,
                  architecture: str,
                  **kwargs):
    system = canonicalize_sys(system)
    architecture = canonicalize_arch(architecture)

    os = f_os(system)
    arch = f_arch(architecture)

    configs = {}
    configs.update(kwargs)

    configs.update({
        "system": system,
        "System": f_System(system),
        "SYSTEM": f_SYSTEM(system),

        "sys": f_sys(system),
        "Sys": f_Sys(system),
        "SYS": f_SYS(system),

        "os": os,
        "Os": f_Os(system),
        "OS": f_OS(system),

        "architecture": architecture,

        "Arch": f_Arch(architecture),
        "arch": arch,
        "ARCH": f_ARCH(architecture),

        "osarch": f_osarch(os, architecture),
        "Osarch": f_Osarch(os, architecture),
        "OSarch": f_OSarch(os, architecture),

        "osbit": f_osbit(os, architecture),

        "bit": f_bit(architecture),
        "extension": f_extension(system),
        "build": f_build(plain_version),
        "version": f_version(plain_version),
        "major_version": f_major_version(plain_version),
        "vmajor_version": f_vmajor_version(plain_version),
        "minor_version": f_minor_version(plain_version),
        "vminor_version": f_vminor_version(plain_version),
        "patch_version": f_patch_version(plain_version),
        "vpatch_version": f_vpatch_version(plain_version)
    })

    configs.update({
        "filename": f_filename(**configs),
        "latest_filename": f_latest_filename(**configs)
    })

    return configs



from .utils.defaults import default_symlink_dir
from .utils import current_system
from .utils import color
from .install import get_exec_version

import os
import shutil
import pathlib
import re


def create_symlink(symlink_path, target_path):
    """
        Make symlink at path `symlink_path` so that it points to `target_path`.
    """
    if current_system() == "winnt":
        os.remove(symlink_path)
        if target_path.endswith('.cmd'):
            shutil.copy(target_path, symlink_path)
        with open(symlink_path, 'w') as f:
            # create a cmd file to mimic how we do symlinks in linux
            f.writelines(['@echo off\n', f'"{target_path}" %*'])
    else:
        target_path = str(pathlib.Path(target_path).resolve())
        if os.path.exists(symlink_path):
            os.remove(symlink_path)
        os.symlink(target_path, symlink_path)


def is_symlink(file):
    if current_system() == "winnt":
        return file.endswith("cmd")
    else:
        return os.path.islink(file)


def get_juliafile_from_version(version):
    if version in ['latest', 'dev']:
        return version
    if re.match('^\d+(.\d+)?$', version):  # <major> or <major>.<minor>
        return f'julia-{version}'
    elif re.match('^\d+.\d+.\d+$', version):
        print(f"{color.YELLOW}Patch version is currently ignored.{color.END}")  # nopep8
        version = '.'.join(version.split('.')[0:2])
        return f'julia-{version}'
    else:
        return None


def switch_julia_target(version_or_path, *,
                        target='julia',
                        symlink_dir=None):
    """
        Switch the julia target version or path.

    Arguments:
      version:
        A path or version of the new julia executable. For version input format is "<major>" or
        "<major>.<minor>".
        For ambiguious input such as "1.10", which will be interpreted as float number 1.1,
        you can either do `jill switch '"1.10"'` or `jill switch 1.10.0`.
      target: (Optional)
        By default it is 'julia', you can also use other target names (e.g., 'julia-1').
      symlink_dir: (Optional)
        The symlink dir that `jill list` looks at.
    """
    symlink_dir = symlink_dir if symlink_dir else default_symlink_dir()
    if not os.path.exists(symlink_dir):
        raise(ValueError(f"symlink dir {symlink_dir} doesn't exist."))

    julia_symlink = os.path.join(symlink_dir, target)
    if not os.path.exists(julia_symlink):
        print(f"{color.RED}{julia_symlink} doesn't exist.{color.END}")
        return False
    if not is_symlink(julia_symlink):
        print(f"{color.RED}Found a suspicious situation: the current {julia_symlink} is not a symlink.{color.END}")
        return False

    version_or_path = str(version_or_path)
    if os.path.exists(version_or_path):
        target_julia = version_or_path
    else:
        version = version_or_path
        target_filename = get_juliafile_from_version(version)
        if not target_filename:
            print(f"{color.RED}Unrecognized input {version}{color.END}")
            return False

        if current_system() == "winnt":
            # JILL uses .cmd to simulate simlinks
            target_filename += ".cmd"
        target_julia = os.path.join(symlink_dir, target_filename)
    if not os.path.exists(target_julia):
        print(
            f"{color.RED}Target {target_julia} doesn't exist.{color.END}")
        return False

    cur_version = get_exec_version(julia_symlink)
    cur_version = f"{color.YELLOW}Invalid{color.END}" if cur_version == "0.0.1" else cur_version
    target_version = get_exec_version(target_julia)
    if target_version == "0.0.1":
        print(f"{color.RED}Target file {target_julia} is invalid.{color.END}")  # nopep8
        return False

    create_symlink(julia_symlink, target_julia)
    print(f"The {color.UNDERLINE}{target}{color.END} target has been changed from {color.UNDERLINE}{cur_version}{color.END} to {color.UNDERLINE}{target_version}{color.END}")  # nopep8
    return True



from .utils.defaults import default_depot_path, default_install_dir, default_symlink_dir
from .utils.filters import f_major_version, f_minor_version
from .utils import query_yes_no
from .utils import current_architecture, current_system, current_libc
from .utils import latest_version
from .utils import DmgMounter, TarMounter
from .utils import Version
from .utils import verify_upstream
from .utils import color, show_verbose
from .download import download_package

import os
import re
import shutil
import subprocess


def is_installed(version, check_symlinks=True):
    """
        check if the required version is already installed.
    """
    check_list = ["julia"]
    if version == "latest":
        check_list.append("julia-latest")
    if version != "latest" and check_symlinks:
        check_list.extend([f"julia-{f_major_version(version)}",
                           f"julia-{f_minor_version(version)}"])

    for path in check_list:
        if Version(get_exec_version(shutil.which(path))) != Version(version):
            return False
    return True


def get_exec_version(path):
    ver_cmd = [path, "--version"]
    try:
        # outputs: "julia version 1.4.0-rc1"
        version = subprocess.check_output(ver_cmd).decode("utf-8")
        version = version.lower().split("version")[-1].strip()
    except:  # nopep8
        # in case it fails in any situation: invalid target or command(.cmd)
        # issue: https://github.com/abelsiqueira/jill/issues/25
        version = "0.0.1"
    return version


def check_installer(installer_path, ext):
    filename = os.path.basename(installer_path)
    if not filename.endswith(ext):
        msg = f"The installer {filename} should be {ext} file"
        raise ValueError(msg)


def last_julia_version(version=None):
    # version should follow semantic version syntax
    def sort_key(ver):
        return float(ver.lstrip("v"))

    version = float(f_minor_version(version)) if version else 999.999
    proj_versions = os.listdir(os.path.join(default_depot_path(),
                                            "environments"))
    proj_versions = [x for x in proj_versions if re.fullmatch(r"v\d+\.\d+", x)]
    proj_versions = sorted(filter(lambda ver: sort_key(ver) < version,
                                  proj_versions),
                           key=sort_key)
    if proj_versions:
        return proj_versions[-1]
    else:
        return None


def make_symlinks(src_bin, symlink_dir, version):
    if not os.path.isfile(src_bin):
        raise(ValueError(f"{src_bin} doesn't exist."))

    system = current_system()
    if symlink_dir not in map(os.path.normpath, os.environ["PATH"].split(os.pathsep)):
        print(f"add {symlink_dir} to PATH")
        if system == "winnt":
            # FIXME: this alse copies system PATH to user PATH
            subprocess.run(["powershell.exe",
                            "setx", "PATH", f'"$env:PATH;{symlink_dir}"'])
        else:
            msg = "~/.bashrc will be modified"
            msg += "\nif you're not using BASH, then you'll need manually"
            msg += f" add {symlink_dir} to your PATH"
            print(msg)

            rc_file = os.path.expanduser("~/.bashrc")
            with open(rc_file, "a") as file:
                file.writelines("\n# added by jill\n")
                file.writelines(f"export PATH={symlink_dir}:$PATH\n")
        print(f"you need to restart your current shell to update PATH")

    os.makedirs(symlink_dir, exist_ok=True)

    new_ver = Version(get_exec_version(src_bin))
    if version == "latest":
        # issue 11: don't symlink to julia
        link_list = ["julia-latest"]
    elif len(Version(version).build) > 0:
        link_list = ["julia-dev"]
    elif len(new_ver.prerelease) > 0:
        # issue #76
        # - it is usually unwanted to symlink unstable release to `julia` and `julia-x`
        # - still symlink to `julia-x.y` because otherwise there is no way to access the unstable
        #   release.
        link_list = [f"julia-{f_minor_version(version)}"]
    else:
        link_list = [f"julia-{f(version)}" for f in (f_major_version,
                                                     f_minor_version)]
        link_list.append("julia")

    for linkname in link_list:
        linkpath = os.path.join(symlink_dir, linkname)
        if current_system() == "winnt":
            linkpath += ".cmd"
        # symlink rules:
        # 1. always symlink latest
        # 2. only make new symlink if it's a newer version
        #   - julia      --> latest stable X.Y.Z
        #   - julia-1    --> latest stable 1.Y.Z
        #   - julia-1.0  --> latest stable 1.0.Z
        #   - don't make symlink to patch level
        if os.path.exists(linkpath) or os.path.islink(linkpath):
            if (os.path.islink(linkpath) and
                    os.readlink(linkpath) == src_bin):
                # happens when installing a new patch version
                continue

            old_ver = Version(get_exec_version(linkpath))
            if show_verbose():
                print(f"old symlink version: {old_ver}")
                print(f"new installation version: {new_ver}")
            if old_ver > new_ver:
                # if two versions are the same, use the new one
                continue

            msg = f"{color.YELLOW}remove old symlink"
            msg += f" {linkname}{color.END}"
            print(msg)
            os.remove(linkpath)
        print(f"{color.GREEN}make new symlink {linkpath}{color.END}")
        if current_system() == "winnt":
            with open(linkpath, 'w') as f:
                # create a cmd file to mimic how we do symlinks in linux
                f.writelines(['@echo off\n', f'"{src_bin}" %*'])
        else:
            os.symlink(src_bin, linkpath)


def copy_root_project(version):
    mver = f_minor_version(version)
    old_ver = last_julia_version(version)
    if old_ver is None:
        print(
            f"Can't find available old root project for version {version}")
        return None

    env_path = os.path.join(default_depot_path(), "environments")
    src_path = os.path.join(env_path, old_ver)
    dest_path = os.path.join(env_path, f"v{mver}")

    if src_path == dest_path:
        return None

    if os.path.exists(dest_path):
        bak_path = os.path.join(env_path, f"v{mver}.bak")
        if os.path.exists(bak_path):
            print(f"{color.YELLOW}delete old backup {bak_path}{color.END}")
            shutil.rmtree(bak_path)
        shutil.move(dest_path, bak_path)
        print(f"{color.YELLOW}move {dest_path} to {bak_path}{color.END}")
    shutil.copytree(src_path, dest_path)


def install_julia_tarball(package_path,
                          install_dir,
                          symlink_dir,
                          version,
                          upgrade):
    check_installer(package_path, ".tar.gz")

    if re.match("(.*)\+(\w+)$", version):
        # We want a different folder name for commit builds so that we can have
        # julia-dev and julia-latest points to two different julia versions
        suffix = 'dev'
    else:
        suffix = f_minor_version(version)

    with TarMounter(package_path) as root:
        src_path = root
        dest_path = os.path.join(install_dir, f"julia-{suffix}")
        if os.path.exists(dest_path):
            shutil.rmtree(dest_path)
            msg = f"{color.YELLOW}remove previous Julia installation:"
            msg += f" {dest_path}{color.END}"
            print(msg)
        # preserve lib symlinks, otherwise it might cause troubles
        # see also: https://github.com/JuliaGPU/CUDA.jl/issues/249
        shutil.copytree(src_path, dest_path, symlinks=True)
        print(f"{color.GREEN}install Julia to {dest_path}{color.END}")
    os.chmod(dest_path, 0o755)  # issue 12
    bin_path = os.path.join(dest_path, "bin", "julia")
    if current_system() == 'winnt':
        bin_path += '.exe'
    make_symlinks(bin_path, symlink_dir, version)
    if upgrade:
        copy_root_project(version)
    return True


def install_julia_dmg(package_path,
                      install_dir,
                      symlink_dir,
                      version,
                      upgrade):
    check_installer(package_path, ".dmg")

    with DmgMounter(package_path) as root:
        # mounted image contents:
        #   ['.VolumeIcon.icns', 'Applications', 'Julia-1.3.app']
        appname = next(filter(lambda x: x.lower().startswith('julia'),
                              os.listdir(root)))
        src_path = os.path.join(root, appname)
        dest_path = os.path.join(install_dir, appname)
        if os.path.exists(dest_path):
            msg = f"{color.YELLOW}remove previous Julia installation:"
            msg += f" {dest_path}{color.END}"
            print(msg)
            shutil.rmtree(dest_path)
        # preserve lib symlinks, otherwise it might cause troubles
        # see also: https://github.com/JuliaGPU/CUDA.jl/issues/249
        shutil.copytree(src_path, dest_path, symlinks=True)
        print(f"{color.GREEN}install Julia to {dest_path}{color.END}")
    bin_path = os.path.join(dest_path,
                            "Contents", "Resources", "julia", "bin", "julia")
    make_symlinks(bin_path, symlink_dir, version)
    if upgrade:
        copy_root_project(version)
    return True


def install_julia_exe(package_path,
                      install_dir,
                      symlink_dir,
                      version,
                      upgrade):
    check_installer(package_path, ".exe")

    dest_path = os.path.join(install_dir,
                             f"julia-{f_minor_version(version)}")
    if os.path.exists(dest_path):
        shutil.rmtree(dest_path, ignore_errors=True)
        msg = f"{color.YELLOW}remove previous Julia installation:"
        msg += f" {dest_path}{color.END}"
        print(msg)

    # build system changes for windows after 1.4
    # https://github.com/JuliaLang/julia/blob/release-1.4/NEWS.md#build-system-changes
    if Version(version).next_patch() < Version("1.4.0"):
        # it's always false if version == "latest"
        subprocess.check_output([f'{package_path}',
                                 '/S', f'/D={dest_path}'])
    else:
        subprocess.check_output([f'{package_path}',
                                 '/VERYSILENT',
                                 f'/DIR={dest_path}'])
    print(f"{color.GREEN}install Julia to {dest_path}{color.END}")
    bin_path = os.path.join(dest_path, "bin", "julia.exe")
    make_symlinks(bin_path, symlink_dir, version)
    if upgrade:
        copy_root_project(version)
    return True


def hello_msg():
    msg = f"{color.BOLD}JILL - Julia Installer 4 Linux"
    msg += f" (MacOS, Windows and FreeBSD) -- Light{color.END}\n"
    print(msg)


def install_julia(version=None, *,
                  install_dir=None,
                  symlink_dir=None,
                  upgrade=False,
                  upstream=None,
                  unstable=False,
                  keep_downloads=False,
                  confirm=False,
                  reinstall=False):
    """
    Install the Julia programming language for your current system

    `jill install [version]` would satisfy most of your use cases, try it first
    and then read description of other arguments. `version` is optional, valid
    version syntax for it is:

    * `stable`: latest stable Julia release. This is the _default_ option.
    * `1`: latest `1.y.z` Julia release.
    * `1.0`: latest `1.0.z` Julia release.
    * `1.4.0-rc1`: as it is.
    * `latest`/`nightly`: the nightly builds from source code.

    For Linux/FreeBSD systems, if you run this command with `root` account,
    then it will install Julia system-widely.

    To download from a private mirror, please check `jill download -h`.

    Arguments:
      version:
        The Julia version you want to install.
      upstream:
        manually choose a download upstream. For example, set it to "Official"
        if you want to download from JuliaComputing's s3 buckets.
      upgrade:
        add `--upgrade` flag also copy the root environment from an older
        Julia version.
      unstable:
        add `--unstable` flag to allow installation of unstable releases for auto version
        query. For example, `jill install --unstable` might give you unstable installation
        like `1.7.0-beta1`. Note that if you explicitly pass the unstable version, e.g.,
        `jill install 1.7.0-beta1`, it will still work.
      keep_downloads:
        add `--keep_downloads` flag to not remove downloaded releases.
      confirm: add `--confirm` flag to skip interactive prompt.
      reinstall:
        jill will skip the installation if the required Julia version already exists,
        add `--reinstall` flag to force the reinstallation.
      install_dir:
        where you want julia packages installed.
      symlink_dir:
        where you want symlinks(e.g., `julia`, `julia-1`) placed.
    """
    install_dir = install_dir if install_dir else default_install_dir()
    install_dir = os.path.abspath(install_dir)
    symlink_dir = symlink_dir if symlink_dir else default_symlink_dir()
    symlink_dir = os.path.normpath(os.path.abspath(symlink_dir))
    system, arch = current_system(), current_architecture()
    version = str(version) if (version or str(version) == "0") else ''
    version = "latest" if version == "nightly" else version
    version = "" if version == "stable" else version
    upstream = upstream if upstream else os.environ.get("JILL_UPSTREAM", None)

    if system == "linux" and current_libc() == "musl":
        # currently Julia tags musl as a system, e.g.,
        # https://julialang-s3.julialang.org/bin/musl/x64/1.5/julia-1.5.1-musl-x86_64.tar.gz
        system = "musl"

    hello_msg()
    if system == "winnt":
        install_dir = install_dir.replace("\\\\", "\\").strip('\'"')
    if not confirm:
        version_str = version if version else "latest stable release"
        question = "jill will:\n"
        question += f"  1) install Julia {version_str} for {system}-{arch}"
        question += f" into {color.UNDERLINE}{install_dir}{color.END}\n"
        question += f"  2) make symlinks in {color.UNDERLINE}{symlink_dir}{color.END}\n"
        question += f"You may need to manually add {color.UNDERLINE}{symlink_dir}{color.END} to PATH\n"
        question += "Continue installation?"
        to_continue = query_yes_no(question)
        if not to_continue:
            return False

    if upstream:
        verify_upstream(upstream)
    wrong_args = False
    try:
        version = latest_version(
            version, system, arch, upstream=upstream, stable_only=not unstable)
    except ValueError:
        # hide the nested error stack :P
        wrong_args = True
    if wrong_args:
        msg = f"wrong version(>= 0.6.0) argument: {version}\n"
        msg += f"Example: `jill install 1`"
        raise(ValueError(msg))

    if not reinstall and is_installed(version):
        print(f"julia {version} already installed.")
        return True

    overwrite = True if version == "latest" else False
    print(f"{color.BOLD}----- Download Julia -----{color.END}")
    package_path = download_package(version, system, arch,
                                    upstream=upstream,
                                    overwrite=overwrite)
    if not package_path:
        return False

    if package_path.endswith(".dmg"):
        installer = install_julia_dmg
    elif package_path.endswith(".tar.gz"):
        installer = install_julia_tarball
    elif package_path.endswith(".exe"):
        installer = install_julia_exe
    else:
        print(f"{color.RED}Unsupported file format for {package_path}{color.END}.")

    print(f"{color.BOLD}----- Install Julia -----{color.END}")
    installer(package_path, install_dir, symlink_dir, version, upgrade)

    if not keep_downloads:
        print(f"{color.BOLD}----- Post Installation -----{color.END}")
        print("remove downloaded files...")
        print(f"remove {package_path}")
        os.remove(package_path)
        gpg_signature_file = package_path + ".asc"
        if os.path.exists(gpg_signature_file):
            print(f"remove {gpg_signature_file}")
            os.remove(gpg_signature_file)
    print(f"{color.GREEN}Done!{color.END}")



from .utils.defaults import default_symlink_dir
from .utils import color
from .install import get_exec_version

import re
import os
import subprocess


def search_files(dir, pattern):
    files = os.listdir(dir)
    julias = [x for x in files if re.match(pattern, x)]
    return julias


def get_julia_build(path):
    ver_cmd = [path, "--startup=no",
               "-e", "println(VERSION, ',', Base.GIT_VERSION_INFO.commit_short)"]
    try:
        build = subprocess.check_output(ver_cmd).decode("utf-8")
        ver, build = build.lower().split(',')
        return ver.strip(), build.strip()
    except:
        return "", ""


def get_binversion(path):
    binversion = get_exec_version(path)
    if binversion.endswith('dev'):
        binversion, build = get_julia_build(path)
    else:
        # we want quick response and we don't really care the commit version of a stable release
        build = ''
    return binversion, build


def sorted_display_list(julialist, reverse):
    # We sort the julia paths in the following order for better display:
    #   julia
    #   julia-<major>
    #   julia-<major>.<minor>
    #   julia-<special name>

    # I just get lazy to write this up so 🤷‍♂️ as long as it works I'm happy
    julia = []
    julia_major = []
    julia_major_minor = []
    julia_special = []
    for x in julialist:
        if x == 'julia':
            julia.append(x)
        elif re.match(r"^julia.\d+$", x):
            julia_major.append(x)
        elif re.match(r"^julia.\d+\.\d+$", x):
            julia_major_minor.append(x)
        else:
            julia_special.append(x)
    julia.sort(reverse=reverse)  # this actually only have 1 element.
    julia_major.sort(reverse=reverse)
    julia_major_minor.sort(reverse=reverse)
    julia_special.sort(reverse=reverse)
    return [*julia, *julia_major, *julia_major_minor, *julia_special]


def list_julia(version=None, *,
               symlink_dir=None):
    """
        List all Julia executable versions in symlink dir

    Arguments:
      version: (Optional)
        The specific version prefix that you are interested in. For example, `jill list 1` checks
        every symlink that matches the name pattern `^julia-1`. (`julia` is excluded in this case.)
      symlink_dir: (Optional)
        The symlink dir that `jill list` looks at.
    """
    symlink_dir = symlink_dir if symlink_dir else default_symlink_dir()
    if not os.path.exists(symlink_dir):
        raise(ValueError(f"symlink dir {symlink_dir} doesn't exist!"))

    version = str(version) if version else ''
    if version:
        pattern = f"^julia-{version}"
    else:
        pattern = "^julia"
    julias = search_files(symlink_dir, pattern=pattern)
    julias = sorted_display_list(julias, reverse=True)
    version_list = [get_binversion(os.path.join(symlink_dir, x)) for x in julias]  # nopep8

    print(f"Found {len(julias)} julia(s) in {color.UNDERLINE}{symlink_dir}{color.END}:")  # nopep8
    for binname, (binversion, build) in zip(julias, version_list):
        if binname.endswith(".cmd"):
            # hide .cmd ext for Windows
            binname = os.path.splitext(binname)[0]
        build_msg = f"+{build}" if build else ""
        binversion_msg = f"{color.RED}Invalid{color.END}" if binversion == '0.0.1' else binversion
        print(f"{color.BOLD}{binname:12}{color.END}\t-->\t{binversion_msg}{build_msg}")
